/*
 * Created on Nov 4, 2004
 */
package org.tekkotsu.ui.editor.netcode;

import java.beans.PropertyChangeSupport;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.Socket;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;
import org.tekkotsu.ui.rcp.Global;
import org.tekkotsu.ui.util.Debugger;

/**
 * @author asangpet
 */
public class RobotConnector {
	public static final String EVENT_DATA_UPDATE = "_model_conn_update";
	public static final String EVENT_MODEL_UPDATE = "_model_conn_model_update";
	public static final String EVENT_TRACE_UPDATE = "_model_conn_trace_update";
	public static final String EVENT_TRACE_UPDATE_TIME = "_model_conn_trace_update_time";
	public static final String EVENT_TRACE_CLEAR = "_model_conn_trace_clear";
	public static final String EVENT_TRACE_MODE_ENTER = "_model_conn_trace_mode";
	public static final String EVENT_TRACE_MODE_EXIT = "_model_conn_trace_exit";

	public static final String defaultHostName    = "localhost";
	public static final String defaultMachineName = "<Name of State Machine>";
	public static final int defaultPort = 10080;

	private int    hostPort    = defaultPort;
	private String hostName;
	private String machineName = Global.startMachineName;

	public PropertyChangeSupport props = new PropertyChangeSupport(this);

	private String command = "find";
	private String lastCommand;

	private boolean commandChange = true;

	public boolean connect() {
		try {
			if (socket != null) {
				disconnect();
			}

			props.firePropertyChange(EVENT_DATA_UPDATE, null, "Connecting to "
					+ getHostName() + ":" + getHostPort() + "...");

			socket = new Socket(hostName, hostPort);
			socketOut = new PrintWriter(socket.getOutputStream(), true);
			socketIn = new BufferedReader(new InputStreamReader(socket
					.getInputStream()));
			props.firePropertyChange(EVENT_DATA_UPDATE, null,
					"Connected Sucessfully");
			return true;
		} catch (java.net.UnknownHostException e) {
			props.firePropertyChange(EVENT_DATA_UPDATE, null, e.toString());
			Debugger.printThrowable(e);

			String msg = e.getMessage();
			MessageDialog.openError(PlatformUI.getWorkbench().getDisplay()
					.getActiveShell(), "Problem", msg);

			return false;
		} catch (Exception e) {
			Debugger.printThrowable(e);
			props.firePropertyChange(EVENT_DATA_UPDATE, null,
					"Unable to connect, please try again");
			props.firePropertyChange(EVENT_DATA_UPDATE, null, e.toString());
			return false;
		}
	}

	public void disconnect() {
		try {
			if (thread != null && thread.isAlive()) {
				props.firePropertyChange(EVENT_DATA_UPDATE, null, "Disconnect");
			}
			stop();
			if (socketOut != null) {
				socketOut.close();
				socketOut = null;
			}
			if (socketIn != null) {
				socketIn.close();
				socketIn = null;
			}
			if (socket != null) {
				socket.close();
				socket = null;
			}
		} catch (Exception e) {
			Debugger.printThrowable(e);
		}
	}

	public void reconnect() {
		stop();
		disconnect();
		connect();
	}

	private Socket socket = null;
	private BufferedReader socketIn = null;
	private PrintWriter socketOut = null;

	public void start() {
		if (socket == null) {
			if (!connect()) {
				return;
			}
		}

		stop();
		execute("listen");
		tick.start();
	}

	public void stop() {
		if (thread != null) {
			thread.running = false;
			thread = null;
		}
		tick.stop();
		lastCommand = command;
	}

	public void resume() {
		tick.resume();
		execute(lastCommand);
	}

	// send a query to the robot
	public void execute(String command) {
		setCommand(command);

		if (thread == null || !thread.running) {
			thread = new ConnectorThread();
			thread.start();
		}
	}

	private TimerTick tick = new TimerTick(this);

	private ConnectorThread thread;

	public RobotConnector() {
		this(Global.startHostName,Global.startMachineName,defaultPort);
		Global.startMachineName = null;
	}

	public RobotConnector(String hostname, String machineName, int port) {
		setHostName(hostname);
		setHostPort(port);
	}

	/**
	 * @return Returns the hostName.
	 */
	public String getHostName() {
		return hostName;
	}

	/**
	 * @param hostName
	 *            The hostName to set.
	 */
	public void setHostName(String hostName) {
		if( hostName == null )
			hostName = "";
		if( hostName.isEmpty() )
			hostName = defaultHostName;
		
		this.hostName        = hostName;
		Global.startHostName = hostName;
	}

	/**
	 * @return Returns the hostPort.
	 */
	public int getHostPort() {
		return hostPort;
	}

	/**
	 * @param hostPort
	 *            The hostPort to set.
	 */
	public void setHostPort(int hostPort) {
		this.hostPort = hostPort;
	}

	public String getCommand() {
		return command;
	}

	public void setCommand(String command) {
		this.command = command;
		commandChange = true;
		reconnect();
	}

	public String getMachineName() {
		return machineName;
	}

	public void setMachineName(String machName) {
		if( machName == null )
			machName = "";
		if( machName.isEmpty() )
			machName = defaultMachineName;
		
		this.machineName = machName;
	}

	private class ConnectorThread extends Thread {
		public boolean running = true;
		private StringBuffer outbuf = new StringBuffer();
		private String result = "";

		@Override
		public void run() {
			while (running) {
				if (commandChange) {
					_changeCommand();
				}

				result = null;
				try {
					if( socketIn.ready() ) {
						// Read a line (blocking)
						result = socketIn.readLine();
						props.firePropertyChange(EVENT_DATA_UPDATE, null, result);
					} else {
						//thread has no input, wait 500ms
						try {
							Thread.sleep(500);
						} catch (InterruptedException e) {
							//do nothing
						}
						
						// and try again if we are still running
						if( !running ) {
							return;
						}
						continue;
					}
				} catch (IOException e) {
					Debugger.printThrowable(e);
					continue;
				}

				outbuf.append(result + "\n");
				if (result.indexOf("</model>") > -1) {
					_updateModel();
				} else if (result.indexOf("</event>") > -1) {
					_updateEvent();
				}
			} // end while running
		}

		private void _updateEvent() {
			while (outbuf.toString().indexOf("</event>") > -1) {
				if (running) {
					if (outbuf.toString().indexOf("<event ") > -1) {
						String bufstr = outbuf.toString();
						int idstart = bufstr.indexOf("<event ");
						int idend = bufstr.indexOf("/>");
						if (idend > -1) {
							int idnend = bufstr.indexOf(">");
							if (!(idnend > -1 && idnend < idend)) {
								props.firePropertyChange(EVENT_TRACE_UPDATE,
										null, bufstr.substring(idstart, idend
												+ "/>".length()));
								outbuf.delete(0, idend + ("/>".length()));
							}
						}
					}
					props.firePropertyChange(EVENT_TRACE_UPDATE, null, trimBuf(
							"<event", "</event>", outbuf));
					int idend = outbuf.indexOf("</event>");
					outbuf.delete(0, idend + "</event>".length());
				}
			}
			outbuf = new StringBuffer();
		}

		private String trimBuf(String prefix, String suffix, StringBuffer buf) {
			int idstart = buf.indexOf(prefix);
			int idend = buf.indexOf(suffix);
			if (idstart == -1)
				return prefix + ">" + suffix;
			return buf.substring(idstart, idend + suffix.length());
		}
		
		private void _updateModel() {
			if (running) {
				props.firePropertyChange(EVENT_MODEL_UPDATE, null, trimBuf(
						"<model", "</model>", outbuf));
			}
			outbuf = new StringBuffer();
		}

		/**
		 * If our command changed, send a "clear" request to the robot to stop
		 * sending us the old information.
		 * 
		 * Then request our new command.
		 */
		private void _changeCommand() {
			if( socket == null ) {
				connect();
			}
			
			socketOut.println("clear");
			socketOut.println(getCommand() + " " + machineName); // spider

			// MachineName
			props.firePropertyChange(EVENT_DATA_UPDATE, null,
					"Enter reading loop \"" + getCommand() + "\" "
							+ machineName + " from " + getHostName() + "\r\n");
			result = "";
			outbuf = new StringBuffer();
			commandChange = false;
		}
	}

	public void sync(int time) {
		tick.sync(time);

		props.firePropertyChange(RobotConnector.EVENT_TRACE_UPDATE_TIME,
			null, Integer.valueOf(time));
	}
}