package org.tekkotsu.ui.storyboard.panels;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.InputEvent;
import org.eclipse.draw2d.LayoutManager;
import org.eclipse.draw2d.MouseEvent;
import org.eclipse.draw2d.MouseListener;
import org.eclipse.draw2d.MouseMotionListener;
import org.eclipse.draw2d.Panel;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.jface.resource.ColorRegistry;
import org.tekkotsu.ui.storyboard.ResourceRegistry;
import org.tekkotsu.ui.storyboard.objects.AbstractViewObject;
import org.tekkotsu.ui.storyboard.views.StoryboardViewer;

public abstract class AbstractTimePanel extends Panel implements PropertyChangeListener {
	public static final float timeScale = 0.001f;
	public static final int gap = 8;
	public static final int halfgap = 4;
	public static final int xoffset = 20;
	public static final int VIEW_BUFFER_WIDTH = 300;

	public static final String TIMELINE_CURSOR_CHANGED = "_time_change";

	private int cursor;
	private int traceTime;
	private TimeMarker marker;
	protected StoryboardViewer viewer;
	protected ColorRegistry colorReg;

	AbstractTimePanel(Panel parent, StoryboardViewer _viewer, TimeMarker marker) {
		this.viewer = _viewer;
		this.marker = marker;
		colorReg = ResourceRegistry.getInstance().getColorRegistry();

		addMouseListener(new MouseListener.Stub() {
			@Override
			public void mousePressed(MouseEvent me) {
				if (!viewer.traceRunning)
					if (me.button == 1)
						moveCursorToOffset(me.getLocation().x);
			}
		});

		addMouseMotionListener(new MouseMotionListener.Stub() {
			@Override
			public void mouseDragged(MouseEvent me) {
				if (!viewer.traceRunning)
					if (me.getState() == InputEvent.BUTTON1)
						moveCursorToOffset(me.getLocation().x);
			}
		});

		parent.add(this);
	}

	@Override
	public void addNotify() {
		super.addNotify();

		setLayoutManager(getParent().getLayoutManager());
		setConstraint(getPreferredX(), getPreferredY(), getPreferredW(),
				getPreferredH());
	}

	protected int getPreferredX() {
		return 0;
	}

	protected int getPreferredY() {
		return 0;
	}

	protected int getPreferredW() {
		return viewer.getVisibleViewport().width;
	}

	protected int getPreferredH() {
		return viewer.getVisibleViewport().height - getPreferredY();
	}

	public int getOffsetOfTime(int time) {
		return marker.getOffset(time);
	}

	public int getTimeOfOffset(int offset) {
		return marker.getTime(offset);
	}

	public TimeMarker getMarker() {
		return marker;
	}

	public Rectangle setConstraint(int x, int y, int w, int h) {
		// ensure always wide enough to fill the window
		w = Math.max(w, viewer.getVisibleViewport().width - x);

		Rectangle rect = new Rectangle(x, y, w, h);
		getParent().getLayoutManager().setConstraint(this, rect);
		return rect;
	}

	public Rectangle getConstraint() {
		IFigure parent = getParent();
		LayoutManager layout = parent.getLayoutManager();
		Rectangle bounds = (Rectangle) layout.getConstraint(this);
		return bounds.getCopy();
	}

	public int getTraceTime() {
		return traceTime;
	}

	public void setTraceTime(int time) {
		traceTime = time;
		ensureTimeLimit(time);
	}

	public int getTimeCursor() {
		return cursor;
	}

	public void setTimeCursor(int i) {
		if (cursor == i) {
			return;
		}

		cursor = i;
		if (cursor < 0) {
			cursor = 0;
		}
		timeCursorUpdated();
	}

	public void timeCursorUpdated() {
		repaint();
	}

	public void ensureTimeLimit(int time) {		
		final Rectangle BOUNDS = getConstraint();
		final int LIMIT = getOffsetOfTime(BOUNDS.width);
		final int NEED = Math.max(getOffsetOfTime(time) + VIEW_BUFFER_WIDTH,
				getPreferredW());

		if (NEED > LIMIT) {
			setConstraint(BOUNDS.x, BOUNDS.y, NEED, BOUNDS.height);
		}

		if( viewer.traceRunning ) {
			moveCursorToTime(time);
			viewer.pushVisibleViewportRight(BOUNDS.x + NEED);
		}
	}

	public void moveCursorToOffset(int x) {
		moveCursorToTime(getTimeOfOffset(x));
	}

	public void moveCursorToTime(int time) {
		setTimeCursor(time);
		firePropertyChange(TIMELINE_CURSOR_CHANGED, null, Integer
				.valueOf(getTimeCursor()));
	}

	public void propertyChange(PropertyChangeEvent evt) {
		if (evt.getPropertyName().equals(TIMELINE_CURSOR_CHANGED)) {
			setTimeCursor(((Integer) evt.getNewValue()).intValue());
		}
	}

	public void reset() {
		if( !viewer.traceRunning ) {
			setTimeCursor(0);
			setTraceTime(0);
		}
		marker.clearMarker();
		setConstraint(getPreferredX(), getPreferredY(), getPreferredW(),
				getPreferredH());
		repaint();
	}

	abstract public void update(AbstractViewObject child);
}
